@@ -16,22 +16,25 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 # 
 # author Salvo "LtWorf" Tomaselli <tiposchi@tiscali.it>
-
-'''This module contains functions to perform various optimizations on the expression trees.
-The list general_optimizations contains pointers to general functions, so they can be called
-within a cycle.
-
-It is possible to add new general optimizations by adding the function in the list
-general_optimizations present in this module. And the optimization will be executed with the
-other ones when optimizing.
-
-A function will have one parameter, which is the root node of the tree describing the expression.
-The class used is defined in optimizer module.
-A function will have to return the number of changes performed on the tree.
-'''
+#
+# This module contains functions to perform various optimizations on the expression trees.
+# The list general_optimizations contains pointers to general functions, so they can be called
+# within a cycle.
+#
+# It is possible to add new general optimizations by adding the function in the list
+# general_optimizations present in this module. And the optimization will be executed with the
+# other ones when optimizing.
+#
+# A function will have one parameter, which is the root node of the tree describing the expression.
+# The class used is defined in optimizer module.
+# A function will have to return the number of changes performed on the tree.
 
 import optimizer
 import parser
+
+from cStringIO import StringIO
+from tokenize import generate_tokens
+
 sel_op=('//=','**=','and','not','in','//','**','<<','>>','==','!=','>=','<=','+=','-=','*=','/=','%=','or','+','-','*','/','&','|','^','~','<','>','%','=','(',')',',','[',']')
 
 def replace_node(replace,replacement):
@@ -48,7 +51,17 @@ def replace_node(replace,replacement):
         replace.left=replacement.left
 
 def recoursive_scan(function,node,rels=None):
-    '''Does a recoursive optimization on the tree'''
+    '''Does a recoursive optimization on the tree.
+    
+    This function will recoursively execute the function given
+    as "function" parameter starting from node to all the tree.
+    if rels is provided it will be passed as argument to the function.
+    Otherwise the function will be called just on the node.
+    
+    Result value: function is supposed to return the amount of changes
+    it has performed on the tree.
+    The various result will be added up and this final value will be the
+    returned value.'''
     changes=0
     #recoursive scan
     if node.kind==optimizer.UNARY:
@@ -68,11 +81,20 @@ def recoursive_scan(function,node,rels=None):
 
 def duplicated_select(n):
     '''This function locates and deletes things like
-    σ a ( σ a(C)) and the ones like σ a ( σ b(C))'''
+    σ a ( σ a(C)) and the ones like σ a ( σ b(C))
+    replacing the 1st one with a single select and
+    the 2nd one with a single select with both conditions
+    in and
+    '''
     changes=0
     if n.name=='σ' and n.child.name=='σ':        
         if n.prop != n.child.prop: #Nested but different, joining them
             n.prop = n.prop + " and " + n.child.prop
+            
+            #This adds parenthesis if they are needed
+            if n.child.prop.startswith('(') or n.prop.startswith('('):
+                n.prop= '(%s)' % n.prop
+                
         n.child=n.child.child
         changes=1
         changes+=duplicated_select(n)
@@ -92,28 +114,38 @@ def futile_union_intersection_subtraction(n):
     
     changes=0
     
+    #Union and intersection of the same thing
     if n.name in ('ᑌ','ᑎ') and n.left==n.right:
         changes=1
         replace_node(n,n.left)
-    elif (n.name == 'ᑌ' and ((n.left.name=='σ' and n.left.child==n.right) or (n.right.name=='σ' and n.right.child==n.left))): #Union of two equal things, but one has a selection
-        changes=1
-        if n.left=='σ':#Selection on left. replacing self with right.
+    
+    #selection and union of the same thing
+    elif (n.name == 'ᑌ'):
+        if n.left.name=='σ' and n.left.child==n.right:
+            changes=1
             replace_node(n,n.right)
-        else:#Selection on left. replacing self with right.
+        elif n.right.name=='σ' and n.right.child==n.left:
+            changes=1
             replace_node(n,n.left)
-    elif (n.name == 'ᑎ' and ((n.left.name=='σ' and n.left.child==n.right) or (n.right.name=='σ' and n.right.child==n.left))): #Intersection of two equal things, but one has a selection
-        changes=1
-        if n.left=='σ':#Swapping with the selection
+            
+    #selection and intersection of the same thing
+    elif n.name == 'ᑎ':
+        if n.left.name=='σ' and n.left.child==n.right:
+            changes=1
             replace_node(n,n.left)
-        else:
+        elif n.right.name=='σ' and n.right.child==n.left:
+            changes=1
             replace_node(n,n.right)
-    #TODO make work the following line...
+            
+    #Subtraction and selection of the same thing
     elif (n.name == '-' and (n.right.name=='σ' and n.right.child==n.left)): #Subtraction of two equal things, but one has a selection
         n.name=n.right.name
         n.kind=n.right.kind
         n.child=n.right.child
         n.prop='(not (%s))' % n.right.prop
         n.left=n.right=None
+    
+    #Subtraction of the same thing or with selection on the left child
     elif (n.name=='-' and ((n.left==n.right) or (n.left.name=='σ' and n.left.child==n.right)) ):#Empty relation
         changes=1
         n.kind=optimizer.UNARY
@@ -307,58 +339,33 @@ def tokenize_select(expression):
     selection. The expression can contain parenthesis.
     It will use a subclass of str with the attribute level, which
     will specify the nesting level of the token into parenthesis.'''
-    sel_op=('//=','**=','and ','not ','in ','//','**','<<','>>','==','!=','>=','<=','+=','-=','*=','/=','%=','or ','+','-','*','/','&','|','^','~','<','>','%','=','(',')',',','[',']')
-    l=0
-    while l!=len(expression):
-        l=len(expression)
-        if expression.startswith('(') and parser.find_matching_parenthesis(expression)+1==len(expression):
-            expression= expression[1:-1]
+    g=generate_tokens(StringIO(expression).readline)
+    l=list(token[1] for token in g)
+    
+    l.remove('')
+    
+    #Changes the 'a','.','method' token group into a single 'a.method' token
+    try:
+        while True:
+            dot=l.index('.')
+            l[dot]='%s.%s' % (l[dot-1],l[dot+1])
+            l.pop(dot+1)
+            l.pop(dot-1)
+    except:
+        pass
+    
     
-    tokens=[]
-    temp=''
     level=0
-    
-    while len(expression)!=0:
-        expression=expression.strip()
+    for i in range(len(l)):
+        l[i]=level_string(l[i])
+        l[i].level=level
         
-        if expression[0:1]=='(': #Expression into parenthesis
+        if l[i]=='(':
             level+=1
-        elif expression[0:1]==')':
+        elif l[i]==')':
             level-=1
-        
-        for i in range(4,0,-1):#operators
-            if expression[0:i] in sel_op:
-                t=level_string(temp)
-                t.level=level
-                tokens.append(t)
-                temp=''
-                t=level_string(expression[0:i].strip())
-                t.level=level
-                tokens.append(t)
-                expression=expression[i:]
-        if expression[0:1]=="'":#String
-            end=expression.index("'",1)
-            while expression[end-1]=='\\':
-                end=expression.index("'",end+1)
-            #Add string to list
-            t=level_string(expression[0:end+1])
-            t.level=level
-            tokens.append(t)
-            expression=expression[end+1:]
-        else:
-            temp+=expression[0:1]
-            expression=expression[1:]
-            pass
-    if len(temp)!=0:
-        t=level_string(temp)
-        t.level=level
-        tokens.append(t)
-    while True:
-        try:
-            tokens.remove('')
-        except:
-            break
-    return tokens
+    
+    return l
 
 def swap_rename_projection(n):
     '''This function locates things like π k(ρ j(R))
@@ -441,6 +448,27 @@ def swap_rename_select(n):
     
     return changes+recoursive_scan(swap_rename_select,n)
 
+def select_union_intersect_subtract(n):
+    '''This function locates things like σ i(a) ᑌ σ q(a)
+    and replaces them with σ (i OR q) (a)
+    Removing a O(n²) operation like the union'''
+    changes=0
+    if n.name in ('ᑌ', 'ᑎ', '-') and n.left.name=='σ' and n.right.name=='σ' and n.left.child==n.right.child:
+        cahnges=1
+        
+        d={'ᑌ':'or', 'ᑎ':'and', '-':'and not'}
+        op=d[n.name]
+        
+        newnode=parser.node()
+        
+        newnode.prop='((%s) %s (%s))' % (n.left.prop,op,n.right.prop)
+        newnode.name='σ'
+        newnode.child=n.left.child
+        newnode.kind=parser.UNARY
+        replace_node(n,newnode)
+        
+    return changes+recoursive_scan(select_union_intersect_subtract,n)
+
 def selection_and_product(n,rels):
     '''This function locates things like σ k (R*Q) and converts them into
     σ l (σ j (R) * σ i (Q)). Where j contains only attributes belonging to R,
@@ -541,7 +569,7 @@ def selection_and_product(n,rels):
             
     return changes+recoursive_scan(selection_and_product,n,rels)
         
-general_optimizations=[duplicated_select,down_to_unions_subtractions_intersections,duplicated_projection,selection_inside_projection,subsequent_renames,swap_rename_select,futile_union_intersection_subtraction,swap_union_renames,swap_rename_projection]
+general_optimizations=[duplicated_select,down_to_unions_subtractions_intersections,duplicated_projection,selection_inside_projection,subsequent_renames,swap_rename_select,futile_union_intersection_subtraction,swap_union_renames,swap_rename_projection,select_union_intersect_subtract]
 specific_optimizations=[selection_and_product]
 
 if __name__=="__main__":
