@@ -17,15 +17,26 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 # 
 # author Salvo "LtWorf" Tomaselli <tiposchi@tiscali.it>
+#
+#
+#
+# This module implements a parser for relational algebra, and can be used
+# to convert expressions into python expressions and to get the parse-tree
+# of the expression.
+#
+# The input must be provided in UTF-8
+# 
+# Language definition here:
+# https://galileo.dmi.unict.it/wiki/relational/doku.php?id=language
 
 
 RELATION=0
 UNARY=1
 BINARY=2
-b_operators=('*','-','ᑌ','ᑎ','÷','ᐅᐊ','ᐅLEFTᐊ','ᐅRIGHTᐊ','ᐅFULLᐊ')
-u_operators=('π','σ','ρ')
+b_operators=('*','-','ᑌ','ᑎ','÷','ᐅᐊ','ᐅLEFTᐊ','ᐅRIGHTᐊ','ᐅFULLᐊ') # List of binary operators
+u_operators=('π','σ','ρ') # List of unary operators
 
-op_functions={'*':'product','-':'difference','ᑌ':'union','ᑎ':'intersection','÷':'division','ᐅᐊ':'join','ᐅLEFTᐊ':'outer_left','ᐅRIGHTᐊ':'outer_right','ᐅFULLᐊ':'outer','π':'projection','σ':'selection','ρ':'rename'}
+op_functions={'*':'product','-':'difference','ᑌ':'union','ᑎ':'intersection','÷':'division','ᐅᐊ':'join','ᐅLEFTᐊ':'outer_left','ᐅRIGHTᐊ':'outer_right','ᐅFULLᐊ':'outer','π':'projection','σ':'selection','ρ':'rename'} # Associates operator with python method
 
 class node (object):
     '''This class is a node of a relational expression. Leaves are relations and internal nodes are operations.
@@ -40,6 +51,7 @@ class node (object):
     
     This class is used to convert an expression into python code.'''
     kind=None
+    __hash__=None
     
     def __init__(self,expression=None):
         '''Generates the tree from the tokenized expression
@@ -106,7 +118,7 @@ class node (object):
             return self.name
         pass
     def printtree(self,level=0):
-        '''Prints a representation of the tree using indentation'''
+        '''returns a representation of the tree using indentation'''
         r=''
         for i in range(level):
             r+='  '
@@ -120,7 +132,7 @@ class node (object):
             
         return '\n'+r
     def get_left_leaf(self):
-        '''This function returns the most left leaf in the tree. It is needed by some optimizations.'''
+        '''This function returns the leftmost leaf in the tree. It is needed by some optimizations.'''
         if self.kind==RELATION:
             return self
         elif self.kind==UNARY:
@@ -131,8 +143,8 @@ class node (object):
         
     def result_format(self,rels):
         '''This function returns a list containing the fields that the resulting relation will have.
-        Since it needs to know real instances of relations, it requires a dictionary where keys are
-        the names of the relations and the values are the relation objects.'''
+        It requires a dictionary where keys are the names of the relations and the values are
+        the relation objects.'''
         if rels==None:            
             return
         
@@ -175,6 +187,7 @@ class node (object):
         if self.kind==BINARY:
             return self.left==other.left and self.right==other.right
         return True
+        
     def __str__(self):
         if (self.kind==RELATION):
             return self.name
@@ -192,15 +205,15 @@ class node (object):
                 
             return (le+ self.name +re)
 
-def find_matching_parenthesis(expression,start=0):
+def _find_matching_parenthesis(expression,start=0,openpar=u'(',closepar=u')'):
     '''This function returns the position of the matching
     close parenthesis to the 1st open parenthesis found
     starting from start (0 by default)'''
     par_count=0 #Count of parenthesis
     for i in range(start,len(expression)):
-        if expression[i]=='(':
+        if expression[i]==openpar:
             par_count+=1
-        elif expression[i]==')':
+        elif expression[i]==closepar:
             par_count-=1
             if par_count==0:
                 return i #Closing parenthesis of the parameter
@@ -235,7 +248,7 @@ def tokenize(expression):
     while len(expression)>0:
         if expression.startswith('('): #Parenthesis state
             state=2
-            end=find_matching_parenthesis(expression)
+            end=_find_matching_parenthesis(expression)
             #Appends the tokenization of the content of the parenthesis
             items.append(tokenize(expression[1:end]))
             #Removes the entire parentesis and content from the expression
@@ -246,7 +259,7 @@ def tokenize(expression):
             expression=expression[2:].strip() #Removing operator from the expression
             
             if expression.startswith('('): #Expression with parenthesis, so adding what's between open and close without tokenization
-                par=expression.find('(',find_matching_parenthesis(expression)) 
+                par=expression.find('(',_find_matching_parenthesis(expression)) 
             else: #Expression without parenthesis, so adding what's between start and parenthesis as whole  
                 par=expression.find('(')
             
@@ -309,7 +322,7 @@ def parse(expr):
     
     You can use parenthesis to change priority: a ᐅᐊ (q ᑌ d).
     
-    IMPORTANT: The encoding used by this module is UTF-8
+    IMPORTANT: The encoding used by this module is UTF-8 (all strings must be UTF-8)
     
     EXAMPLES
     σage > 25 and rank == weight(A)
