@@ -16,6 +16,9 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 # 
 # author Salvo "LtWorf" Tomaselli <tiposchi@tiscali.it>
+#
+# This module provides a classes to represent relations and to perform
+# relational operations on them.
 
 from rtypes import *
 import csv
@@ -25,67 +28,62 @@ class relation (object):
     A relation can be represented using a table
     Calling an operation and providing a non relation parameter when it is expected will
     result in a None value'''    
-    def __init__(self,filename="",comma_separated=True):
+    __hash__=None
+    
+    def __init__(self,filename=""):
         '''Creates a relation, accepts a filename and then it will load the relation from
         that file. If no parameter is supplied an empty relation is created. Empty
         relations are used in internal operations.
         By default the file will be handled like a comma separated as described in
         RFC4180, but it can also be handled like a space separated file (previous
         default format) setting to false the 2nd parameter.
-        The old format is deprecated since it doesn't permit fields
-        with spaces, you should avoid using it.'''
+        The old format is no longer supported.'''
+        
+        self._readonly=False
+        
         if len(filename)==0:#Empty relation
-            self.content=[]
+            self.content=set()
             self.header=header([])
             return
         #Opening file
         fp=file(filename)
-        if comma_separated:
-            reader=csv.reader(fp) #Creating a csv reader
-            self.header=header(reader.next()) # read 1st line
-            self.content=[]
-            for i in reader.__iter__(): #Iterating rows
-                self.content.append(i)
-        else: #Old format
-            self.header=header(fp.readline().replace("\n","").strip().split(" "))
+
+        reader=csv.reader(fp) #Creating a csv reader
+        self.header=header(reader.next()) # read 1st line
+        self.content=set()
         
-            self.content=[]
-            row=fp.readline()
-            while len(row)!=0:#Reads the content of the relation
-                self.content.append(row.replace("\n","").strip().split(" "))
-                row=fp.readline()
+        for i in reader.__iter__(): #Iterating rows
+            self.content.add(tuple(i))
         
         #Closing file
         fp.close()
         
-    
-    def save(self,filename,comma_separated=True):
+    def _make_writable(self):
+        '''If this relation is marked as readonly, this 
+        method will copy the content to make it writable too'''
+        
+        if self._readonly:
+            self.content=set(self.content)
+            self._readonly=False
+        
+    def save(self,filename):
         '''Saves the relation in a file. By default will save using the csv
         format as defined in RFC4180, but setting comma_separated to False,
         it will use the old format with space separated values.
-        The old format is deprecated since it doesn't permit fields
-        with spaces, you should avoid using it.'''
+        The old format is no longer supported.'''
         
         fp=file(filename,'w') #Opening file in write mode
-        if comma_separated: #writing csv
-            writer=csv.writer(fp) #Creating csv writer
+        
+        writer=csv.writer(fp) #Creating csv writer
             
-            #It wants an iterable containing iterables
-            head=[]
-            head.append(self.header.attributes)
-            writer.writerows(head)
+        #It wants an iterable containing iterables
+        head=(self.header.attributes,)
+        writer.writerows(head)
             
-            #Writing content, already in the correct format
-            writer.writerows(self.content)
-        else: #Writing in the old, deprecated, format
-            res=""
-            res+=" ".join(self.header.attributes)
-            
-            for r in self.content:
-                res+="\n"
-                res+=" ".join(r)
-            fp.write(res)
+        #Writing content, already in the correct format
+        writer.writerows(self.content)
         fp.close() #Closing file
+                
     def _rearrange_(self,other):
         '''If two relations share the same attributes in a different order, this method
         will use projection to make them have the same attributes' order.
@@ -97,7 +95,20 @@ class relation (object):
         if self.header.sharedAttributes(other.header) == len(self.header.attributes) == len(other.header.attributes):
             return other.projection(list(self.header.attributes))
         return None
-        
+    
+    def _autocast(self,string):
+        '''Depending on the regexp matched by the string,
+        it will perform automatic casting'''
+        tmpstring=rstring(string)
+        if len(tmpstring)>0 and tmpstring.isInt():
+            return int(tmpstring)
+        elif len(tmpstring)>0 and tmpstring.isFloat():
+            return float(tmpstring)
+        elif len(tmpstring)>0 and tmpstring.isDate():
+            return rdate(tmpstring)
+        else:
+            return tmpstring
+    
     def selection(self,expr):
         '''Selection, expr must be a valid boolean expression, can contain field names,
         constant, math operations and boolean ones.'''
@@ -105,19 +116,13 @@ class relation (object):
         newt=relation()
         newt.header=header(list(self.header.attributes))
         for i in self.content:
+	    #Fills the attributes dictionary with the values of the tuple
             for j in range(len(self.header.attributes)):
-                if len(i[j])>0 and i[j].isdigit():
-                    attributes[self.header.attributes[j]]=int(i[j])
-                elif len(i[j])>0 and rstring(i[j]).isFloat():
-                    attributes[self.header.attributes[j]]=float(i[j])
-                elif len(i[j])>0 and isDate(i[j]):
-                    attributes[self.header.attributes[j]]=rdate(i[j])
-                else:
-                    attributes[self.header.attributes[j]]=i[j]
+                attributes[self.header.attributes[j]]=self._autocast(i[j])
                 
             try:
                 if eval(expr,attributes):
-                    newt.content.append(i)
+                    newt.content.add(i)
             except Exception,e:
                 raise Exception("Failed to evaluate %s\n%s" % (expr,e.__str__()))
         return newt
@@ -134,7 +139,7 @@ class relation (object):
         
         for i in self.content:
             for j in other.content:
-                newt.content.append(i+j)
+                newt.content.add(i+j)
         return newt
         
     
@@ -174,8 +179,7 @@ class relation (object):
             row=[]
             for j in ids:
                 row.append(i[j])
-            if attributes_same_count or row not in newt.content:
-                newt.content.append(row)
+            newt.content.add(tuple(row))
         return newt
     
     def rename(self,params):
@@ -192,7 +196,8 @@ class relation (object):
             if (newt.header.rename(old,new)) == False:
                 return None
         
-        newt.content=list(self.content)
+        newt.content=self.content
+        newt._readonly=True
         return newt
         
     def intersection(self,other):
@@ -207,10 +212,7 @@ class relation (object):
         newt=relation()
         newt.header=header(list(self.header.attributes))
         
-        #Adds only element not in other, duplicating them
-        for e in self.content:
-            if e in other.content:
-                newt.content.append(list(e))
+        newt.content=self.content.intersection(other.content)
         return newt
     
     def difference(self,other):
@@ -225,10 +227,7 @@ class relation (object):
         newt=relation()
         newt.header=header(list(self.header.attributes))
         
-        #Adds only element not in other, duplicating them
-        for e in self.content:
-            if e not in other.content:
-                newt.content.append(list(e))
+        newt.content=self.content.difference(other.content)
         return newt
     def division(self,other):
         '''Division operator
@@ -272,13 +271,7 @@ class relation (object):
         newt=relation()
         newt.header=header(list(self.header.attributes))
         
-        #Adds element from self, duplicating them all
-        for e in self.content:
-            newt.content.append(list(e))
-        
-        for e in other.content:
-            if e not in newt.content:
-                newt.content.append(list(e))
+        newt.content=self.content.union(other.content)
         return newt
     def thetajoin(self,other,expr):
         '''Defined as product and then selection with the given expression.'''
@@ -344,34 +337,32 @@ class relation (object):
                     for l in noid:
                         item.append(j[l])
                     
-                    newt.content.append(item)
+                    newt.content.add(tuple(item))
                     added=True
             #If it didn't partecipate, adds it
             if not added:
                 item=list(i)
                 for l in range(len(noid)):
                     item.append("---")
-                newt.content.append(item)
+                newt.content.add(tuple(item))
         
         return newt
     
     def join(self,other):
         '''Natural join, joins on shared attributes (one or more). If there are no
         shared attributes, it will behave as cartesian product.'''
-        shared=[]
-        for i in self.header.attributes:
-            if i in other.header.attributes:
-                shared.append(i)
+        
+        #List of attributes in common between the relations
+        shared=list(set(self.header.attributes).intersection(set(other.header.attributes)))
         
         newt=relation() #Creates the new relation
         
-        #Adds all the attributes of the 1st relation
+        #Adding to the headers all the fields, done like that because order is needed
         newt.header=header(list(self.header.attributes))
-        
-        #Adds all the attributes of the 2nd, when non shared
         for i in other.header.attributes:
             if i not in shared:
                 newt.header.attributes.append(i)
+                
         #Shared ids of self
         sid=self.header.getAttributesId(shared)
         #Shared ids of the other relation
@@ -394,30 +385,24 @@ class relation (object):
                     for l in noid:
                         item.append(j[l])
                     
-                    newt.content.append(item)
+                    newt.content.add(tuple(item))
         
         return newt
     def __eq__(self,other):
         '''Returns true if the relations are the same, ignoring order of items.
         This operation is rather heavy, since it requires sorting and comparing.'''
         other=self._rearrange_(other) #Rearranges attributes' order so can compare tuples directly
+        
         if (self.__class__!=other.__class__)or(self.header!=other.header):
             return False #Both parameters must be a relation
+
+        if set(self.header.attributes)!=set(other.header.attributes):
+            return False
+        
         
-        #Comparing header
-        if len(self.header.attributes) != len(other.header.attributes):
-            return False #Not the same number of attributes -> not equals
-        for i in self.header.attributes:
-            if i not in other.header.attributes:
-                return False #Non shared attribute
         
         #comparing content
-        if len(self.content) != len(other.content):
-            return False #Not the same 
-        for i in self.content:
-            if i not in other.content:
-                return False
-        return True
+        return self.content==other.content
         
     def __str__(self):
         '''Returns a string representation of the relation, can be printed with 
@@ -433,7 +418,6 @@ class relation (object):
                     m_len[col]=len(i)
                 col+=1
                 
-        
         res=""
         for f in range(len(self.header.attributes)):
             res+="%s"%(self.header.attributes[f].ljust(2+m_len[f]))
@@ -456,6 +440,7 @@ class relation (object):
         Dic must be a dictionary that has the form field name:value. Every kind of value
         will be converted into a string.
         Returns the number of affected rows.'''
+        self._make_writable()
         affected=0
         attributes={}
         keys=dic.keys() #List of headers to modify
@@ -464,15 +449,8 @@ class relation (object):
         #new_content=[] #New content of the relation
         for i in self.content:
             for j in range(len(self.header.attributes)):
-                #Giving to the field it's right format (hopefully)
-                if i[j].isdigit():
-                    attributes[self.header.attributes[j]]=int(i[j])
-                elif rstring(i[j]).isFloat():
-                    attributes[self.header.attributes[j]]=float(i[j])
-                elif isDate(i[j]):
-                    attributes[self.header.attributes[j]]=rdate(i[j])
-                else:
-                    attributes[self.header.attributes[j]]=i[j]
+                attributes[self.header.attributes[j]]=self._autocast(i[j])
+                
             if eval(expr,attributes): #If expr is true, changing the tuple
                 affected+=1
                 new_tuple=list(i)
@@ -482,8 +460,9 @@ class relation (object):
                 self.content.remove(i) 
                 for k in range(len(keys)):
                     new_tuple[f_ids[k]]=str(dic[keys[k]])
-                self.content.append(new_tuple)
+                self.content.add(tuple(new_tuple))
         return affected
+        
     def insert(self,values):
         '''Inserts a tuple in the relation.
         This function will not insert duplicate tuples.
@@ -494,16 +473,16 @@ class relation (object):
         if len(self.header.attributes) != len(values):
             return 0
             
+        self._make_writable()
+            
         #Creating list containing only strings
         t=[]
         for i in values:
             t.append(str(i))
         
-        if t not in self.content:
-            self.content.append(t)
-            return 1
-        else:
-            return 0
+        prevlen=len(self.content)
+        self.content.add(tuple(t))
+        return len(self.content)-prevlen
     
     def delete(self,expr):
         '''Delete, expr must be a valid boolean expression, can contain field names,
@@ -511,22 +490,18 @@ class relation (object):
         This operation will change the relation itself instead of generating a new one,
         deleting all the tuples that make expr true.
         Returns the number of affected rows.'''
+        self._make_writable()
         attributes={}
         affected=len(self.content)
-        new_content=[] #New content of the relation
+        new_content=set() #New content of the relation
         for i in self.content:
             for j in range(len(self.header.attributes)):
-                if i[j].isdigit():
-                    attributes[self.header.attributes[j]]=int(i[j])
-                elif rstring(i[j]).isFloat():
-                    attributes[self.header.attributes[j]]=float(i[j])
-                elif isDate(i[j]):
-                    attributes[self.header.attributes[j]]=rdate(i[j])
-                else:
-                    attributes[self.header.attributes[j]]=i[j]
+                attributes[self.header.attributes[j]]=self._autocast(i[j])
+                
+                
             if not eval(expr,attributes):
                 affected-=1
-                new_content.append(i)
+                new_content.add(i)
         self.content=new_content
         return affected
     
@@ -534,9 +509,13 @@ class header (object):
     '''This class defines the header of a relation.
     It is used within relations to know if requested operations are accepted'''
     
+    #Since relations are mutalbe we explicitly block hashing them
+    __hash__=None
+    
     def __init__(self,attributes):
         '''Accepts a list with attributes' names. Names MUST be unique'''
         self.attributes=attributes
+        
     def __repr__(self):
         return "header(%s)" % (self.attributes.__repr__())
         
@@ -544,20 +523,16 @@ class header (object):
     def rename(self,old,new):
         '''Renames a field. Doesn't check if it is a duplicate.
         Returns True if the field was renamed, False otherwise'''
-        for i in range(len(self.attributes)):
-            if self.attributes[i]==old:
-                self.attributes[i]=new
-                return True
-        return False #Requested field was not found    
-        
+        try:
+            id_=self.attributes.index(old)
+            self.attributes[id_]=new
+        except:
+            return False
+        return True
     
     def sharedAttributes(self,other):
         '''Returns how many attributes this header has in common with a given one'''
-        res=0
-        for i in self.attributes:
-            if i in other.attributes:
-                res+=1
-        return res
+        return len(set(self.attributes).intersection(set(other.attributes)))
     
     def __str__(self):
         '''Returns String representation of the field's list'''
